//go:build !ignore_autogenerated
// +build !ignore_autogenerated

// Code generated by pointuh. DO NOT EDIT.

package simple

// ExportedInt is an autogenerated clone of simple.ExportedInt
type ExportedInt int

// ExportedString is an autogenerated clone of simple.ExportedString
type ExportedString string

// ExportedStruct is an autogenerated type.
type ExportedStruct struct {
	I  *int
	PI *int
	S  *string
	PS *string
	AI *int
	AS *simple.UnderlyingStruct
}

// UnderlyingStruct is an autogenerated type.
type UnderlyingStruct struct {
}
