// Copyright 2019 The Go Authors. All rights reserved.
// Use of this source code is governed by a BSD-style
// license that can be found in the LICENSE file.

// This file is only here to not trigger "build constraints exclude all Go files" during tests

package testdata
