// Copyright 2022 The Go Authors. All rights reserved.
// Use of this source code is governed by a BSD-style
// license that can be found in the LICENSE file.

package a

import "testing"

var fuzzTargets = []testing.InternalFuzzTarget{
	{"Fuzz", Fuzz},
}

func Fuzz(f *testing.F) {}
