// Copyright 2015 go-swagger maintainers
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//    http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

package spec

import (
	"testing"
)

func TestIntegrationExternalDocs(t *testing.T) {
	var extDocs = ExternalDocumentation{Description: "the name", URL: "the url"}
	const extDocsJSON = `{"description":"the name","url":"the url"}`
	assertSerializeJSON(t, extDocs, extDocsJSON)
	assertParsesJSON(t, extDocsJSON, extDocs)
}
